;;; lurk.el --- Little Unified iRc Klient -*- lexical-binding:t -*-

;; Copyright (C) 2021 Tim Vaughan

;; Author: Tim Vaughan <timv@ughan.xyz>
;; Created: 14 June 2021
;; Version: 1.0
;; Keywords: network
;; Homepage: http://thelambdalab.xyz/lurk
;; Package-Requires: ((emacs "26"))

;; This file is not part of GNU Emacs.

;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with this file.  If not, see <http://www.gnu.org/licenses/>.

;;; Commentary:

;;; Code:

(provide 'lurk)


;;; Customizations
;;

(defgroup lurk nil
  "Little Unified iRc Klient."
  :group 'network)

(defcustom lurk-nick "plugd"
  "Default nick.")

(defcustom lurk-default-quit-msg "Bye"
  "Default quit message when none supplied.")

(defcustom lurk-networks
  '(("libera" "irc.libera.chat" 6697)
    ("freenode" "chat.freenode.net" 6697)
    ("tilde" "tilde.chat" 6697)
    ("mbr" "mbrserver.com" 6667 :notls)
    ("local" "localhost" 6697))
  "IRC networks.")

(defcustom lurk-allow-ipv6 nil
  "Set to non-nil to allow use of IPv6.")

(defcustom lurk-show-joins nil
  "Set to non-nil to be notified of joins, parts and quits.")

;;; Faces
;;

(defface lurk-text
  '((t :inherit font-lock-preprocessor-face))
  "Face used for Lurk text.")

(defface lurk-prompt
  '((t :inherit org-level-2))
  "Face used for the prompt.")

(defface lurk-context
  '((t :inherit org-list-dt))
  "Face used for the context name in the prompt.")

(defface lurk-faded
  '((t :inherit font-lock-preprocessor-face))
  "Face used for faded Lurk text.")

(defface lurk-bold
  '((t :inherit font-lock-function-name-face))
  "Face used for bold Lurk text.")

(defface lurk-error
  '((t :inherit font-lock-regexp-grouping-construct))
  "Face used for Lurk error text.")

;;; Global variables
;;

(defvar lurk-version "Lurk v0.1")

(defvar lurk-notice-prefix "-!-")

(defvar lurk-error-prefix "!!!")

(defvar lurk-prompt-string
  (propertize "> " 'face 'lurk-prompt))


;;; Network process
;;

(defvar lurk-response "")

(defun lurk-filter (proc string)
  (dolist (line (split-string (concat lurk-response string) "\n"))
    (if (string-suffix-p "\r" line)
        (lurk-eval-msg-string (string-trim line))
      (setq lurk-response line))))

(defun lurk-sentinel (proc string)
  (unless (equal "open" (string-trim string))
    (lurk-display-error "Disconnected from server.")
    (clrhash lurk-contexts)
    (setq lurk-current-context nil)
    (lurk-render-prompt)
    (cancel-timer lurk-ping-timer)))

(defun lurk-start-process (network)
  (let* ((row (assoc network lurk-networks))
         (host (elt row 1))
         (port (elt row 2))
         (flags (seq-drop row 3)))
    (make-network-process :name "lurk"
                          :host host
                          :service port
                          :family (if lurk-allow-ipv6 nil 'ipv4)
                          :filter #'lurk-filter
                          :sentinel #'lurk-sentinel
                          :nowait nil
                          :tls-parameters (if (memq :notls flags)
                                              nil
                                            (cons 'gnutls-x509pki
                                                  (gnutls-boot-parameters
                                                   :type 'gnutls-x509pki
                                                   :hostname host)))
                          :buffer "*lurk*")))

(defvar lurk-ping-timer nil)
(defvar lurk-ping-period 60)

(defun lurk-ping-function ()
  (lurk-send-msg (lurk-msg nil nil "PING" (car (process-contact (get-process "lurk")))))
  (setq lurk-ping-timer (run-with-timer lurk-ping-period nil #'lurk-ping-function)))

(defun lurk-connect (network)
  (if (get-process "lurk")
      (lurk-display-error "Already connected.  Disconnect first.")
    (if (not (assoc network lurk-networks))
        (lurk-display-error "Network '" network "' is unknown.")
      (clrhash lurk-contexts)
      (setq lurk-current-context nil)
      (lurk-start-process network)
      (lurk-send-msg (lurk-msg nil nil "USER" lurk-nick 0 "*" lurk-nick))
      (lurk-send-msg (lurk-msg nil nil "NICK" lurk-nick))
      (setq lurk-ping-timer (run-with-timer lurk-ping-period nil #'lurk-ping-function)))))

(defun lurk-connected-p ()
  (let ((proc (get-process "lurk")))
    (and proc (eq (process-status proc) 'open))))

(defun lurk-send-msg (msg)
  (let ((proc (get-process "lurk")))
    (if (and proc (eq (process-status proc) 'open))
        (process-send-string proc (concat (lurk-msg->string msg) "\r\n"))
      (lurk-display-error "No server connection established.")
      (error "No server connection established"))))


;;; Server messages
;;

(defun lurk--as-string (obj)
  (if obj
      (with-output-to-string (princ obj))
    nil))

(defun lurk-msg (tags src cmd &rest params)
  (list (lurk--as-string tags)
        (lurk--as-string src)
        (upcase (lurk--as-string cmd))
        (mapcar #'lurk--as-string
                (if (and params (listp (elt params 0)))
                    (elt params 0)
                  params))))

(defun lurk-msg-tags (msg) (elt msg 0))
(defun lurk-msg-src (msg) (elt msg 1))
(defun lurk-msg-cmd (msg) (elt msg 2))
(defun lurk-msg-params (msg) (elt msg 3))
(defun lurk-msg-trail (msg)
  (let ((params (lurk-msg-params msg)))
    (if params
        (elt params (- (length params) 1)))))

(defvar lurk-msg-regex
  (rx
   (opt (: "@" (group (* (not (or "\n" "\r" ";" " ")))))
        (* whitespace))
   (opt (: ":" (: (group (* (not (any space "!" "@"))))
                  (* (not (any space)))))
        (* whitespace))
   (group (: (* (not whitespace))))
   (* whitespace)
   (opt (group (+ not-newline))))
  "Regex used to parse IRC messages.
Note that this regex is incomplete.  Noteably, we discard the non-nick
portion of the source component of the message, as LURK doesn't use this.")

(defun lurk-string->msg (string)
  (if (string-match lurk-msg-regex string)
      (let* ((tags (match-string 1 string))
             (src (match-string 2 string))
             (cmd (upcase (match-string 3 string)))
             (params-str (match-string 4 string))
             (params
              (if params-str
                  (let* ((idx (cl-search ":" params-str))
                         (l (split-string (string-trim (substring params-str 0 idx))))
                         (r (if idx (list (substring params-str (+ 1 idx))) nil)))
                    (append l r))
                nil)))
        (apply #'lurk-msg (append (list tags src cmd) params)))
    (error "Failed to parse string " string)))

(defun lurk--filtered-join (&rest args)
  (string-join (seq-filter (lambda (el) el) args) " "))

(defun lurk-msg->string (msg)
  (let ((tags (lurk-msg-tags msg))
        (src (lurk-msg-src msg))
        (cmd (lurk-msg-cmd msg))
        (params (lurk-msg-params msg)))
    (lurk--filtered-join
     (if tags (concat "@" tags) nil)
     (if src (concat ":" src) nil)
     cmd
     (if (> (length params) 1)
         (string-join (seq-take params (- (length params) 1)) " ")
       nil)
     (if (> (length params) 0)
         (concat ":" (elt params (- (length params) 1)))
       nil))))


;;; Contexts
;;

(defvar lurk-current-context nil)
(defvar lurk-contexts (make-hash-table :test #'equal))

(defun lurk-add-context (name)
  (puthash name nil lurk-contexts))

(defun lurk-del-context (name)
  (remhash name lurk-contexts))

(defun lurk-get-context-users (name)
  (gethash name lurk-contexts))

(defun lurk-add-context-users (context users)
  (puthash context
           (append users
                   (gethash context lurk-contexts))
           lurk-contexts))

(defun lurk-del-context-user (context user)
  (puthash context
           (remove user (gethash context lurk-contexts))
           lurk-contexts))

(defun lurk-del-user (user)
  (dolist (context (lurk-get-context-list))
    (lurk-del-context-user context user)))

(defun lurk-rename-user (old-nick new-nick)
  (dolist (context (lurk-get-context-list))
    (lurk-del-context-user context old-nick)
    (lurk-add-context-users context (list new-nick))))

(defun lurk-get-context-type (name)
  (cond
   ((string-prefix-p "#" name) 'channel)
   ((string-match-p (rx (or "." "localhost")) name) 'host)
   (t 'nick)))

(defun lurk-get-context-list ()
  (let ((res nil))
    (maphash (lambda (key val)
               (cl-pushnew key res))
             lurk-contexts)
    res))

(defun lurk-get-next-context (&optional prev)
  (if lurk-current-context
      (let* ((context-list (if prev
                               (reverse (lurk-get-context-list))
                             (lurk-get-context-list)))
             (context-list* (member lurk-current-context context-list)))
        (if (> (length context-list*) 1)
            (cadr context-list*)
          (car context-list)))
    nil))

(defun lurk-cycle-contexts (&optional rev)
  (if lurk-current-context
      (progn
        (setq lurk-current-context (lurk-get-next-context rev))
        (lurk-render-prompt))
    (lurk-display-error "No channels joined.")))

(defun lurk-cycle-contexts-forward ()
  (interactive)
  (lurk-cycle-contexts)
  (lurk-highlight-context lurk-current-context))

(defun lurk-cycle-contexts-reverse ()
  (interactive)
  (lurk-cycle-contexts t)
  (lurk-highlight-context lurk-current-context))


;;; Buffer
;;
(defun lurk-render-prompt ()
  (with-current-buffer "*lurk*"
    (let ((update-point (= lurk-input-marker (point)))
          (update-window-points (mapcar (lambda (w)
                                          (list (= (window-point w) lurk-input-marker)
                                                w))
                                        (get-buffer-window-list nil nil t))))
      (save-excursion
        (set-marker-insertion-type lurk-prompt-marker nil)
        (set-marker-insertion-type lurk-input-marker t)
        (let ((inhibit-read-only t))
          (delete-region lurk-prompt-marker lurk-input-marker)
          (goto-char lurk-prompt-marker)
          (insert
           (propertize (if lurk-current-context
                           lurk-current-context
                         "")
                       'face 'lurk-context
                       'read-only t)
           (propertize lurk-prompt-string
                       'face 'lurk-prompt
                       'read-only t
                       'rear-nonsticky t)))
        (set-marker-insertion-type lurk-input-marker nil))
      (if update-point
          (goto-char lurk-input-marker))
      (dolist (v update-window-points)
        (if (car v)
            (set-window-point (cadr v) lurk-input-marker))))))
  
(defvar lurk-prompt-marker nil
  "Marker for prompt position in LURK buffer.")

(defvar lurk-input-marker nil
  "Marker for prompt position in LURK buffer.")

(defun lurk-setup-buffer ()
  (with-current-buffer (get-buffer-create "*lurk*")
    (setq-local scroll-conservatively 1)
    (if (markerp lurk-prompt-marker)
        (set-marker lurk-prompt-marker (point-max))
      (setq lurk-prompt-marker (point-max-marker)))
    (if (markerp lurk-input-marker)
        (set-marker lurk-input-marker (point-max))
      (setq lurk-input-marker (point-max-marker)))
    (goto-char (point-max))
    (lurk-render-prompt)))


;;; Output formatting and highlighting
;;

;; Partially-implemented idea: the face text property can be
;; a list of faces, applied in order.  By assigning each context
;; a unique list and keeping track of these in a hash table, we can
;; easily switch the face corresponding to a particular context
;; by modifying the elements of this list.
;;
;; More subtly, we make only the cdrs of this list shared among
;; all text of a given context, allowing the cars to be different
;; and for different elements of the context-specific text to have
;; different styling.

;; Additionally, we can allow selective hiding of contexts via
;; the buffer-invisibility-spec.

(defvar lurk-context-facelists (make-hash-table :test 'equal)
  "List of seen contexts and associated face lists.")

(defun lurk-get-context-facelist (context)
  (let ((facelist (gethash context lurk-context-facelists)))
    (unless facelist
      (setq facelist (list 'lurk-text))
      (puthash context facelist lurk-context-facelists))
    facelist))

(defun lurk-display-string (context &rest strings)
  (with-current-buffer (get-buffer-create "*lurk*")
    (save-excursion
      (goto-char lurk-prompt-marker)
      (let ((inhibit-read-only t)
            (old-pos (marker-position lurk-prompt-marker))
            (adaptive-fill-regexp (rx (= 6 anychar)))
            (fill-column 80)
            (context-atom (if context (intern context) nil)))
        (insert-before-markers
         (propertize (concat (format-time-string "%H:%M") " ")
                     'face (lurk-get-context-facelist context)
                     'read-only t
                     'context context
                     'invisible context-atom
                     'help-echo (concat "Context: " (or context "none")))
         (propertize (concat (apply #'concat strings) "\n")
                     'face (lurk-get-context-facelist context)
                     'read-only t
                     'context context
                     'invisible context-atom
                     'help-echo (concat "Context: " (or context "none"))))
        (fill-region old-pos lurk-prompt-marker nil t)))))

(defun lurk-display-message (from to text)
  (let ((context (if (eq 'channel (lurk-get-context-type to))
                     to
                   (if (equal to lurk-nick) from to))))
    (lurk-display-string
     context
     (pcase (lurk-get-context-type to)
       ('channel (concat to " <" from "> " text))
       ('nick (concat "[" from " -> " to "] " text))
       (_
        (error "Unsupported context type"))))))

(defun lurk-display-notice (context &rest notices)
  (lurk-display-string
   context
   lurk-notice-prefix " "
   (apply #'concat notices)))

(defun lurk-display-error (&rest messages)
  (lurk-display-string
   nil
   lurk-error-prefix " "
   (apply #'concat messages)))

(defun lurk-highlight-context (context)
  (maphash
   (lambda (this-context facelist)
     (if (equal this-context context)
         (setcar facelist 'lurk-bold)
       (setcar facelist 'lurk-faded)))
   lurk-context-facelists)
  (force-window-update "*lurk*"))

(defun lurk-zoom-in (context)
  (with-current-buffer "*lurk*"
    (maphash
     (lambda (this-context _)
       (let ((this-context-atom (if this-context (intern this-context) nil)))
         (if (equal this-context context)
             (remove-from-invisibility-spec this-context-atom)
           (add-to-invisibility-spec this-context-atom))))
     lurk-context-facelists)
    (force-window-update "*lurk*")))

(defun lurk-zoom-out ()
  (with-current-buffer "*lurk*"
    (maphash
     (lambda (this-context _)
       (let ((this-context-atom (if this-context (intern this-context) nil)))
         (remove-from-invisibility-spec this-context-atom)))
     lurk-context-facelists)
    (force-window-update "*lurk*")))

;;; Message evaluation
;;

(defun lurk-eval-msg-string (string)
  ;; (lurk-display-string nil string)
  (let* ((msg (lurk-string->msg string)))
    ;; (message (pp msg))
    (pcase (lurk-msg-cmd msg)
      ("PING"
       (lurk-send-msg
        (lurk-msg nil nil "PONG" (lurk-msg-params msg))))
       ;; (lurk-display-notice nil "ping-pong (server initiated)"))

      ("PONG")
       ;; (lurk-display-notice nil "ping-pong (client initiated)"))

      ("001"
       (let* ((params (lurk-msg-params msg))
              (nick (elt params 0))
              (text (string-join (seq-drop params 1) " ")))
         (setq lurk-nick nick)
         (lurk-display-notice nil text)))

      ("353" ; NAMEREPLY
       (let* ((params (lurk-msg-params msg))
              (channel (elt params 2))
              (names (split-string (elt params 3))))
         (lurk-add-context-users channel names)))

      ("366" ; ENDOFNAMES
       (let* ((params (lurk-msg-params msg))
              (channel (elt params 1)))
         (lurk-display-notice
          channel
          (lurk--as-string (length (lurk-get-context-users channel)))
          " users in " channel)))

      ((rx (= 3 (any digit)))
       (lurk-display-notice nil (mapconcat 'identity (cdr (lurk-msg-params msg)) " ")))

      ((and "JOIN"
            (guard (equal lurk-nick (lurk-msg-src msg))))
       (let ((channel (car (lurk-msg-params msg))))
         (lurk-add-context channel)
         (setq lurk-current-context channel)
         (lurk-display-notice channel "Joining channel " channel)
         (lurk-render-prompt)))

      ("JOIN"
       (let ((channel (car (lurk-msg-params msg)))
             (nick (lurk-msg-src msg)))
         (lurk-add-context-users channel (list nick))
         (if lurk-show-joins
             (lurk-display-notice channel nick " joined channel " channel))))

      ((and "PART"
            (guard (equal lurk-nick (lurk-msg-src msg))))
       (let ((channel (car (lurk-msg-params msg))))
         (lurk-display-notice channel "Left channel " channel)
         (lurk-del-context channel)
         (if (equal channel lurk-current-context)
             (setq lurk-current-context (lurk-get-next-context)))
         (lurk-render-prompt)))

      ("PART"
       (let ((channel (car (lurk-msg-params msg)))
             (nick (lurk-msg-src msg)))
         (lurk-del-context-user channel nick)
         (if lurk-show-joins
             (lurk-display-notice channel nick " left channel " channel))))

      ("QUIT"
       (let ((nick (lurk-msg-src msg))
             (reason (mapconcat 'identity (lurk-msg-params msg) " ")))
         (lurk-del-user nick)
         (if lurk-show-joins
             (lurk-display-notice nil nick " quit: " reason))))

      ((and "NICK"
            (guard (equal lurk-nick (lurk-msg-src msg))))
       (setq lurk-nick (car (lurk-msg-params msg)))
       (lurk-display-notice nil "Set nick to " lurk-nick))

      ("NICK"
       (let ((old-nick (lurk-msg-src msg))
             (new-nick (car (lurk-msg-params msg))))
         (lurk-display-notice nil old-nick " is now known as " new-nick)
         (lurk-rename-user old-nick new-nick)))

      ("NOTICE"
       (let ((nick (lurk-msg-src msg))
             (channel (car (lurk-msg-params msg)))
             (text (cadr (lurk-msg-params msg))))
         (pcase text
           ((rx (: "\01VERSION "
                   (let version (* (not "\01")))
                   "\01"))
            (lurk-display-notice nil "CTCP version reply from " nick ": " version))
           (_
            (lurk-display-notice nil text)))))

      ("PRIVMSG"
       (let* ((from (lurk-msg-src msg))
              (params (lurk-msg-params msg))
              (to (car params))
              (text (cadr params)))
         (pcase text
           ("\01VERSION\01"
            (let ((version-string (concat lurk-version " - running on GNU Emacs " emacs-version)))
              (lurk-send-msg (lurk-msg nil nil "NOTICE"
                                       (list from (concat "\01VERSION "
                                                          version-string
                                                          "\01")))))
            (lurk-display-notice nil "CTCP version request received from " from))

           ((rx (let ping (: "\01PING " (* (not "\01")) "\01")))
            (lurk-send-msg (lurk-msg nil nil "NOTICE" (list from ping)))
            (lurk-display-notice "CTCP ping received from " from))

           ("\01USERINFO\01"
            (lurk-display-notice "CTCP userinfo request from " from " (no response sent)"))

           (_
            (if (and (equal from "BitBot")
                     (equal to "##moshpit")
                     (cl-search "\\_o< QUACK!" text))
              (lurk-send-msg (lurk-msg nil nil "PRIVMSG" to ",bef")))
            (lurk-display-message from to text)))))
      (_
       (lurk-display-notice nil (lurk-msg->string msg))))))


;;; Command entering
;;

(defun lurk-enter-string (string)
  (if (string-prefix-p "/" string)
      (pcase (substring string 1)
        ((rx (: "CONNECT " (let network (* not-newline))))
         (lurk-display-notice nil "Attempting to connect to " network "...")
         (lurk-connect network))

        ((rx (: "TOPIC " (let new-topic (* not-newline))))
         (lurk-send-msg (lurk-msg nil nil "TOPIC" lurk-current-context new-topic)))

        ((rx (: "ME " (let action (* not-newline))))
         (lurk-send-msg (lurk-msg nil nil "PRIVMSG"
                                  (list lurk-current-context
                                        (concat "\01ACTION " action "\01"))))
         (lurk-display-action lurk-nick action))

        ((rx (: "VERSION" " " (let nick (+ (not whitespace)))))
         (lurk-send-msg (lurk-msg nil nil "PRIVMSG"
                                  (list nick "\01VERSION\01")))
         (lurk-display-notice nil "CTCP version request sent to " nick))

        ((rx "PART" (opt (: " " (let channel (* not-newline)))))
         (if (or lurk-current-context channel)
             (lurk-send-msg (lurk-msg nil nil "PART" (if channel
                                                         channel
                                                       lurk-current-context)))
           (lurk-display-error "No current channel to leave.")))

        ((rx "QUIT" (opt (: " " (let quit-msg (* not-newline)))))
         (lurk-send-msg (lurk-msg nil nil "QUIT"
                                  (or quit-msg lurk-default-quit-msg))))

        ((rx (: "NICK" (* whitespace) string-end))
         (lurk-display-notice nil "Current nick: " lurk-nick))

        ((rx (: "NICK" (+ whitespace) (let nick (+ (not whitespace)))))
         (if (lurk-connected-p)
             (lurk-send-msg (lurk-msg nil nil "NICK" nick))
           (setq lurk-nick nick)
           (lurk-display-notice nil "Set default nick to '" nick "'")))

        ((rx "MSG "
             (let to (* (not whitespace)))
             " "
             (let text (* not-newline)))
         (lurk-send-msg (lurk-msg nil nil "PRIVMSG" to text))
         (lurk-display-message lurk-nick to text))

        ((rx (: (let cmd-str (+ (not whitespace)))
                (opt (: " " (let params-str (* not-newline))))))
         (lurk-send-msg (lurk-msg nil nil (upcase cmd-str)
                                  (if params-str
                                      (split-string params-str)
                                    nil)))))

    (unless (string-empty-p string)
      (if lurk-current-context
          (progn
            (lurk-send-msg (lurk-msg nil nil "PRIVMSG"
                                     lurk-current-context
                                     string))
            (lurk-display-message lurk-nick lurk-current-context string))
        (lurk-display-error "No current context.")))))

(defun lurk-enter ()
  "Enter current contents of line after prompt."
  (interactive)
  (with-current-buffer "*lurk*"
    (let ((line (buffer-substring lurk-input-marker (point-max))))
      (let ((inhibit-read-only t))
        (delete-region lurk-input-marker (point-max)))
      (lurk-enter-string line))))


;;; Mode
;;

(defvar lurk-mode-map
  (let ((map (make-sparse-keymap)))
    (define-key map (kbd "RET") 'lurk-enter)
    (define-key map (kbd "<C-tab>") 'lurk-cycle-contexts-forward)
    (define-key map (kbd "<C-S-tab>") 'lurk-cycle-contexts-reverse)
    map))

(define-derived-mode lurk-mode text-mode "lurk"
  "Major mode for LURK.")

(when (fboundp 'evil-set-initial-state)
  (evil-set-initial-state 'lurk-mode 'insert))

;;; Main start procedure
;;

(defun lurk ()
  "Switch to *lurk* buffer."
  (interactive)
  (if (get-buffer "*lurk*")
      (switch-to-buffer "*lurk*")
    (switch-to-buffer "*lurk*"))
  (lurk-mode)
  (lurk-setup-buffer)
  "Started LURK.")



;;; lurk.el ends here
